using System;
using System.Collections.Generic;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;
using System.Diagnostics;
using System.IO;
using System.Media;
using NativeWifi;

namespace WifiScanner
{
	// A Netstumbler (sort of) for Vista
	// Charles Putney
	// 18 Quinns Road
	// Shankill
	// Co. Dublin
	// IRELAND

	// June 22, 2007 First Release to CodeProject
	// June 23, 2007 Whoops - Bitmaps were loaded from my directory.  Installed bitmaps to listView.
	// August 10, 2007 Update by Norman Rasmussen to add Native Vista API calls

	public partial class Form1 : Form
	{
		private readonly WlanClient.WlanInterface wlanInterface;
		private readonly AutoResetEvent scanComplete = new AutoResetEvent(false);
		private string baseText;

		public Form1()
		{
			InitializeComponent();
			WlanClient wlanClient = new WlanClient();
			wlanInterface = wlanClient.Interfaces[0];
			wlanInterface.WlanNotification += wlanInterface_WlanNotification;
			timer1_Tick(null, null);
		}

		void wlanInterface_WlanNotification(Wlan.WlanNotificationData notifyData)
		{
			if (notifyData.notificationSource == Wlan.WlanNotificationSource.ACM)
			{
				Wlan.WlanNotificationCodeAcm acmCode = (Wlan.WlanNotificationCodeAcm)notifyData.NotificationCode;
				if (acmCode == Wlan.WlanNotificationCodeAcm.ScanComplete || acmCode == Wlan.WlanNotificationCodeAcm.ScanFail)
					scanComplete.Set();
			}
			if (notifyData.notificationSource == Wlan.WlanNotificationSource.MSM)
			{
				Wlan.WlanNotificationCodeMsm msmCode = (Wlan.WlanNotificationCodeMsm)notifyData.NotificationCode;
				if (msmCode == Wlan.WlanNotificationCodeMsm.SignalQualityChange)
				{
					int wlanSignalQuality = 0;
					if (notifyData.dataSize != Marshal.SizeOf(wlanSignalQuality))
						throw new Exception();
					wlanSignalQuality = Marshal.ReadInt32(notifyData.dataPtr, 0);
					Debug.WriteLine(notifyData.interfaceGuid + ": wlanSignalQuality: " + wlanSignalQuality);
				}
			}
			Debug.WriteLine(notifyData.NotificationCode.ToString());
		}

		private DateTime updateStart;
		private void timer1_Tick(object sender, EventArgs e)
		{
			if (baseText == null) baseText = Text;
			timer1.Enabled = false;
			updateStart = DateTime.Now;
			new MethodInvoker(Scan).BeginInvoke(null, null);
		}

		private void Scan()
		{
			List<Network> networks = new List<Network>();

#if !API
			DateTime startTime = DateTime.Now;
			wlanInterface.Scan();
			scanComplete.WaitOne();
			TimeSpan duration = DateTime.Now - startTime;
			Debug.WriteLine("Scanned in " + duration.TotalSeconds.ToString("0.000") + " seconds");
			Dictionary<string, Wlan.WlanAvailableNetwork> wlanNetworks = new Dictionary<string, Wlan.WlanAvailableNetwork>();
			foreach (Wlan.WlanAvailableNetwork network in wlanInterface.AvailableNetworks)
			{
				wlanNetworks[network.dot11Ssid.ToString()] = network;
			}
			Wlan.WlanBssEntry[] bssList = wlanInterface.BssList;
			foreach (Wlan.WlanBssEntry bssEntry in bssList)
			{
				Network network = new Network(bssEntry.dot11Ssid.ToString());
				network.MACAddress = "api:     " + bssEntry.Dot11BssidString;
				network.Channel = ((bssEntry.chCenterFrequency - 2407000) / 5000).ToString();
				int linkQuality = (int)bssEntry.linkQuality;
				if (linkQuality < 0) linkQuality = (linkQuality + 100) * 2; // HACK: convert quality to a percentage
				//if (linkQuality < 0) linkQuality = 100 - (int)(100.0 * (-50.0 - linkQuality) / (-50.0 - -100.0));
				network.Signal = linkQuality + "%";
				Wlan.WlanAvailableNetwork wlanNetwork;
				if (wlanNetworks.TryGetValue(bssEntry.dot11Ssid.ToString(), out wlanNetwork))
				{
					network.Authenticatiopn = wlanNetwork.dot11DefaultAuthAlgorithm.ToString();
					network.Encryption = wlanNetwork.dot11DefaultCipherAlgorithm.ToString();
					network.RadioType = wlanNetwork.dot11PhyTypes[0].ToString();
				}
				network.NetworkType = bssEntry.dot11BssType.ToString();
				double maxRate = 0;
				for (int i = 0; i < bssEntry.wlanRateSet.rateSetLength; i++)
				{
					double rate = (bssEntry.wlanRateSet.usRateSet[i] & 0x7FFF) * 0.5;
					if (rate > maxRate) maxRate = rate;
				}
				network.Speed = maxRate.ToString();
				networks.Add(network);
			}
#endif

#if !NETSH
			string output;
			string line;
			Network currentNetwork = new Network(null);
			int BSSIDNumber = 0;

			Process proc = new Process();
			proc.StartInfo.CreateNoWindow = true;
			proc.StartInfo.FileName = "netsh";
			proc.StartInfo.Arguments = "wlan show networks mode=bssid";
			proc.StartInfo.RedirectStandardOutput = true;
			proc.StartInfo.UseShellExecute = false; // required for the Redirect setting above Process.Start(proc);
			proc.Start();
			output = proc.StandardOutput.ReadToEnd();
			proc.WaitForExit();
			StringReader sr = new StringReader(output);

			while ((line = sr.ReadLine()) != null)
			{
				if (line.StartsWith("General Failure"))
				{
					// Wifi disconnected or not installed
					break;
				}
				if (line.StartsWith("SSID"))
				{
					if (currentNetwork.SSID != null) networks.Add(currentNetwork);
					currentNetwork = new Network(line.Substring(line.IndexOf(":") + 1).Trim());
					currentNetwork.Signal = "0%"; // prevent exception for trim
					BSSIDNumber = 0; // reset the BSSID number
					continue;
				}

				if (line.IndexOf("Network type") > 0)
				{
					if (line.EndsWith("Infrastructure"))
					{
						currentNetwork.NetworkType = "AP";
						continue;
					}
					else
					{
						currentNetwork.NetworkType = line.Substring(line.IndexOf(":") + 1); //"Ad-hoc";
					}
				}
				if (line.IndexOf("Authentication") > 0)
				{
					currentNetwork.Authenticatiopn = line.Substring(line.IndexOf(":") + 1).Trim();
					continue;
				}
				if (line.IndexOf("Encryption") > 0)
				{
					currentNetwork.Encryption = line.Substring(line.IndexOf(":") + 1).Trim();
					continue;
				}
				if (line.IndexOf("BSSID") > 0)
				{
					if ((Convert.ToInt32(line.IndexOf("BSSID" + 6)) > BSSIDNumber))
					{
						networks.Add(currentNetwork);
						BSSIDNumber = Convert.ToInt32(line.IndexOf("BSSID" + 6));
						Network newNetwork = new Network(currentNetwork.SSID);
						newNetwork.NetworkType = currentNetwork.NetworkType;
						newNetwork.Authenticatiopn = currentNetwork.Authenticatiopn;
						newNetwork.Encryption = currentNetwork.Encryption;
						currentNetwork = newNetwork;
					}
					currentNetwork.MACAddress = "netsh: " + line.Substring(line.IndexOf(":") + 1).Trim();
					continue;
				}
				if (line.IndexOf("Signal") > 0)
				{
					currentNetwork.Signal = line.Substring(line.IndexOf(":") + 1);
					continue;
				}
				if (line.IndexOf("Radio Type") > 0)
				{
					currentNetwork.RadioType = line.Substring(line.IndexOf(":") + 1);
					continue;
				}
				if (line.IndexOf("Channel") > 0)
				{
					currentNetwork.Channel = line.Substring(line.IndexOf(":") + 1);
					continue;
				}
				if (line.IndexOf("Basic Rates") > 0)
				{
					//currentNetwork.Speed = line.Substring(line.Length - 2, 2);
					currentNetwork.Speed = line.Substring(line.IndexOf(":"));
					if (currentNetwork.Speed == ":")
					{
						currentNetwork.Speed = "not shown";
						continue;
					}
					currentNetwork.Speed = currentNetwork.Speed.Trim(':', ' ');
					for (int i = currentNetwork.Speed.Length - 1; i > 0; i--)
					{
						if (currentNetwork.Speed.Substring(i, 1) == " ")
						{
							currentNetwork.Speed = currentNetwork.Speed.Substring(i + 1, currentNetwork.Speed.Length - 1 - i);
							break;
						}
					}
				}
				if (line.IndexOf("Other Rates") > 0)
				{
					// overwrite the basic rates if this entry is present
					currentNetwork.Speed = line.Substring(line.IndexOf(":"));
					if (currentNetwork.Speed == ":")
					{
						currentNetwork.Speed = "not shown";
						continue;
					}
					currentNetwork.Speed = currentNetwork.Speed.Trim(':', ' ');
					for (int i = currentNetwork.Speed.Length - 1; i >= 0; i--)
					{
						if (currentNetwork.Speed.Substring(i, 1) == " ")
						{
							currentNetwork.Speed = currentNetwork.Speed.Substring(i + 1, currentNetwork.Speed.Length - 1 - i);
							break;
						}
					}
				}
			}
			if (currentNetwork.SSID != null) networks.Add(currentNetwork);
#endif

			Invoke(new UpdateUIEvent(UpdateUI), networks);
		}

		delegate void UpdateUIEvent(IEnumerable<Network> networks);
		private void UpdateUI(IEnumerable<Network> networks)
		{
			timer1.Enabled = true;
			TimeSpan duration = DateTime.Now - updateStart;
			Text = baseText + " (" + duration.TotalSeconds.ToString("0.000") + " seconds latency)";
			listView1.SmallImageList = imageSignalLevel;
			for (int i = 0; i < listView1.Items.Count; i++)
			{
				// set signal to zero on all items in list
				listView1.Items[i].SubItems[3].Text = "0%";
				listView1.Items[i].ImageIndex = 5;
			}

			foreach (Network network in networks)
			{
				ListViewItem SearchItem;
				if (network.MACAddress.Length == 0) continue; // don't search if no valid MAC Address !
				SearchItem = listView1.FindItemWithText(network.MACAddress);
				if (SearchItem == null)
				{
					// New discovery - add it to the list

					SystemSounds.Hand.Play();

					SearchItem = listView1.Items.Add(network.MACAddress);
					SearchItem.SubItems.Add(network.SSID);
					SearchItem.SubItems.Add(network.Channel);
					SearchItem.SubItems.Add(network.Signal);
					SearchItem.SubItems.Add(network.Authenticatiopn);
					SearchItem.SubItems.Add(network.Encryption);
					SearchItem.SubItems.Add(network.RadioType);
					SearchItem.SubItems.Add(network.NetworkType);
					SearchItem.SubItems.Add(network.Speed);
					SearchItem.EnsureVisible();
				}
				else
				{
					//  Already in list - update Signal and other details that may change

					SearchItem.SubItems[3].Text = network.Signal;

					// Don't change any details if blank 
					if (network.SSID != null) SearchItem.SubItems[1].Text = network.SSID;
					if (network.Authenticatiopn != null) SearchItem.SubItems[4].Text = network.Authenticatiopn;
					if (network.Encryption != null) SearchItem.SubItems[5].Text = network.Encryption;
					if (network.RadioType != null) SearchItem.SubItems[6].Text = network.RadioType;
					if (network.NetworkType != null) SearchItem.SubItems[7].Text = network.NetworkType;
					if (network.Speed != null) SearchItem.SubItems[8].Text = network.Speed;
				}

				int SignalInt = Convert.ToInt32(network.Signal.TrimEnd(' ', '%'));
				if (SignalInt > 50) SearchItem.ImageIndex = 0;
				else if (SignalInt > 40) SearchItem.ImageIndex = 1;
				else if (SignalInt > 30) SearchItem.ImageIndex = 2;
				else if (SignalInt > 20) SearchItem.ImageIndex = 3;
				else if (SignalInt > 0) SearchItem.ImageIndex = 4;

				if (((network.Authenticatiopn == "Open") && (network.Encryption == "None")) ||
					((network.Authenticatiopn == "IEEE80211_Open") && (network.Encryption == "None"))) SearchItem.BackColor = Color.PaleGreen;
			}
		}

		struct Network
		{
			public string MACAddress;
			public readonly string SSID;
			public string Channel;
			public string Signal;
			public string Authenticatiopn;
			public string Encryption;
			public string RadioType;
			public string NetworkType;
			public string Speed;

			public Network(string SSID)
			{
				MACAddress = string.Empty;
				this.SSID = SSID;
				Channel = string.Empty;
				Signal = string.Empty;
				Authenticatiopn = null;
				Encryption = null;
				RadioType = null;
				NetworkType = null;
				Speed = null;
			}
		}
	}
}
